{-# LANGUAGE NoImplicitPrelude #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE DerivingVia #-}
{-# LANGUAGE TupleSections #-}
{-# OPTIONS_GHC -Wno-unrecognised-pragmas #-}
module Chart (newChart, FillerException(..), getUpdatedChartCache) where

import Import
import Data.Aeson (eitherDecodeFileStrict')
import RIO.List
import RIO.List.Partial
import RIO.FilePath
import Data.Time.Calendar (Day(..))
import Data.FingerTree (FingerTree)
import Control.Concurrent (forkIO)
import Control.Monad.Extra (ifM)
import qualified RIO.ByteString as BS
-- import Control.Exception
import qualified Data.HashMap.Strict as HM
import qualified Data.FingerTree as FT
import qualified Data.Text as T

import qualified Debug.Trace as D

data FillerException = QuitFiller
  deriving Show

instance Exception FillerException

--deriving via Integer instance Hashable Day

newChart :: IBContract -> RIO App ()
newChart contract = do
  app <- ask
  let sym = (symbol :: IBContract -> Text) contract
      con = (conId :: IBContract -> Int) contract
      hmVar = appCharts . appRefs $ app
  hm <- liftIO . readTVarIO $ hmVar
  unless (sym `HM.member` hm) $ do
    today <- liftIO getCurrentDay :: RIO App Day
    (cacheErrors, cacheData) <- do
        files <- liftIO $ listDirectory $ "cache" </> show con
        res <- forM files $ \cacheFileName -> do
          let fname = "cache" </> show con </> cacheFileName
          ifM (liftIO $ doesFileExist fname)
              (bimap (\err -> cacheFileName <> ": "<> err) ((readMaybe $ dropExtension cacheFileName) :: Maybe Day,) <$> liftIO (eitherDecodeFileStrict' fname))
              (return $ Left $ "cachefile "<>cacheFileName<>" not found.")
        return $ partitionEithers res
    unless (null cacheErrors) $ forM_ cacheErrors $ \err -> logError (displayShow err)
    let (unknownDates, cacheData') = partition (isNothing . fst) cacheData
    unless (null unknownDates) $ logError $ display $ T.pack $ show (length unknownDates) <> " files had no parsable Date."
    let cacheData'' :: HashMap Day (FingerTree TimePoint ChartPoint) = HM.fromList $ bimap (fromMaybe today) FT.fromList <$> cacheData'
    logError $ displayShow $ HM.keys cacheData''
    c <- liftIO $ newTVarIO $ Chart con (fromMaybe FT.empty $ cacheData'' HM.!? today) (HM.delete today cacheData'') undefined defChartSettings [] Nothing False
    tid <- liftIO $ forkIO $ fillChart app contract c
    liftIO $ atomically $ do
      modifyTVar' c (\Chart{..} -> let fillerThread = tid in Chart{..})
      modifyTVar' hmVar (HM.insert sym c)

fillChart :: App -> IBContract -> TVar Chart -> IO ()
fillChart app contract cVar = runRIO app $ do
  let sym = (symbol :: IBContract -> Text) contract
  (tickerMapVar :: TVar (HashMap Int Text)) <- tickerIdToSymbol. appRefs <$> ask
  alreadyAdded <- (sym `elem`) . HM.elems <$> liftIO (readTVarIO tickerMapVar)
  unless alreadyAdded $ do
    tickerId <- (+1) . foldl' max 1000 . HM.keys <$> liftIO (readTVarIO tickerMapVar)
    let cancelSubscription = liftIO $ atomically $ do
         modifyTVar tickerMapVar (HM.delete tickerId)
         -- TODO: send cancel-request
    let sendQ = twsConnectionSend $ twsConnectionRefs $ appRefs $ app
    liftIO $ atomically $ do
      modifyTVar tickerMapVar (HM.insert tickerId sym)
      writeTQueue sendQ $ Msg_IB_OUT $ IB_RequestMktData tickerId contract "233" False False
    handle (\QuitFiller -> cancelSubscription >> exitSuccess) $
      forever $ do
        -- chart dirty? set clean & begin work
        c <- liftIO (readTVarIO cVar)
        when (chartDirty c) $ do
          liftIO $ atomically $ modifyTVar cVar (\c -> c { chartDirty = False })
          let (lUpdate, cachePoints) = getUpdatedChartCache c Nothing
          liftIO $ atomically $ modifyTVar cVar (\c -> c { chartCache = cachePoints, lastCacheUpdate = lUpdate })
          return ()
        threadDelay 1000000 -- sleep 5 seconds

getUpdatedChartCache :: Chart -> Maybe ChartSettings -> (Maybe TimePoint, [ChartPoint])
getUpdatedChartCache Chart{..} chartSettings' = (lUpdate, takeWhile (\ChartPoint{..} -> (\(TimePoint x) -> x < cacheUpdateStart) timeOfDay) chartCache <> map toCachePoint chunkedChart)
    where (TimePoint cacheUpdateStart) = fromMaybe (TimePoint 0) Nothing -- TODO: lastCacheUpdate
          cacheUpdateEnd   = 86400
          chunkChart :: Int -> Int -> Int -> FingerTree TimePoint ChartPoint -> [(TimePoint,[ChartPoint])]
          chunkChart from until range tree = go from range interval
              where
                lastItem = case FT.viewr interval of
                            FT.EmptyR -> until
                            (_ FT.:> ChartPoint{..}) -> (\(TimePoint x) -> x) timeOfDay
                interval = FT.takeUntil (\(TimePoint x) -> x > until)
                         . FT.dropUntil (\(TimePoint x) -> x > from)
                         $ tree
                go f i t
                  | f+i >= lastItem = [(TimePoint (f+i), toList t)]
                  | otherwise       = let (a, b) = FT.split (\(TimePoint x) -> x > f+i) t
                                      in (TimePoint (f+i),toList a) : go (f+i) i b
          chunkedChart = chunkChart cacheUpdateStart cacheUpdateEnd (chartResolution $ fromMaybe chartSettings chartSettings') chartData
          lUpdate = fmap fst . lastMaybe $ chunkedChart
          toCachePoint :: (TimePoint,[ChartPoint]) -> ChartPoint
          toCachePoint (t,[]) = ChartPoint t (-1) 0 []
          toCachePoint (t,as) = ChartPoint t m vol [OLHC o l h c]
            where
              vol = sum $ volume <$> as
              as' = pointValue <$> as
              ms' = sum $ (\x -> pointValue x * fromIntegral (volume x)) <$> as
              m = ms' / fromIntegral vol
              o = head as'
              c = last as'
              l = minimum as'
              h = maximum as'
